import torch

from invariantkernels import block_permutation_group, permutation_group


def test_permutation_group():
    x = torch.tensor([[1, 2, 3], [4, 5, 6]])
    result = permutation_group(x)
    expected = torch.tensor(
        [
            [[1, 2, 3], [4, 5, 6]],  # Identity
            [[1, 3, 2], [4, 6, 5]],  # Swap last two
            [[2, 1, 3], [5, 4, 6]],  # Swap first two
            [[2, 3, 1], [5, 6, 4]],  # Left shift
            [[3, 1, 2], [6, 4, 5]],  # Right shift
            [[3, 2, 1], [6, 5, 4]],  # swap first and last
        ]
    )
    assert torch.allclose(result, expected)


def test_block_permutation_group():
    x = torch.tensor([[1, 2, 3, 4], [5, 6, 7, 8]])
    result = block_permutation_group(x, 2)
    expected = torch.tensor(
        [
            [[1, 2, 3, 4], [5, 6, 7, 8]],  # Identity
            [[3, 4, 1, 2], [7, 8, 5, 6]],  # Swap blocks
        ]
    )
    assert torch.allclose(result, expected)



import itertools
from typing import Iterator

import torch


def permutation_group(x: torch.Tensor) -> torch.tensor:
    indices = range(x.shape[-1])
    permuted_indices = [list(p) for p in itertools.permutations(indices)]
    permuted_x = x[..., permuted_indices]
    # permuted_x is a tensor of shape (..., n, G, d)
    # Reorder the dimensions to (..., G, n, d)
    dim_indices = list(range(permuted_x.dim()))
    dim_indices[-2], dim_indices[-3] = dim_indices[-3], dim_indices[-2]
    return permuted_x.permute(*dim_indices)


def block_permutation_group(x: torch.Tensor, block_size: int) -> torch.tensor:
    if x.shape[-1] % block_size != 0:
        raise ValueError(
            "Last dimension of x must be a multiple of block size"
            f" (got {x.shape[-1]} and {block_size} respectively)."
        )

    block_indices = [
        range(i, i + block_size) for i in range(0, x.shape[-1], block_size)
    ]
    # Use itertools.chain to flatten the list of lists generated by the permutation
    permuted_indices = [
        list(itertools.chain.from_iterable(p))
        for p in itertools.permutations(block_indices)
    ]
    permuted_x = x[..., permuted_indices]
    # permuted_x is a tensor of shape (..., n, G, d)
    # Reorder the dimensions to (..., G, n, d)
    dim_indices = list(range(permuted_x.dim()))
    dim_indices[-2], dim_indices[-3] = dim_indices[-3], dim_indices[-2]
    return permuted_x.permute(*dim_indices)



from typing import Callable, Iterable, Optional, Tuple

import gpytorch
import torch


class InvariantKernel(gpytorch.kernels.Kernel):
    r"""A kernel that is invariant to a collection of transformations.

    Currently only supports group invariance.

    The invariant kernel is defined as:
    .. math::
        k_G(x, y) = \frac{1}{|G|^2} \sum_{g \in G} \sum_{h \in G} k(g(x), h(y))

    If the kernel is isotropic, we can use the simpler form:
    .. math::
        k_G(x, y) = \frac{1}{|G|} \sum_{g \in G} k(g(x), y)

    where :math:`G` is the group of transformations, :math:`k` is the base kernel, and :math:`x` and :math:`y` are the inputs.
    """

    def __init__(
        self,
        base_kernel: gpytorch.kernels.Kernel,
        transformations: Callable[[torch.tensor], torch.tensor],
        is_isotropic: bool = False,
        is_group: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.base_kernel = base_kernel
        self.transformations = transformations
        self.is_isotropic = is_isotropic
        self.is_group = is_group

        if not self.is_group:
            raise NotImplementedError("InvariantKernel only supports group invariance.")

    def forward(
        self, x1, x2, diag: bool = False, last_dim_is_batch: bool = False, **kwargs
    ) -> torch.tensor:
        if last_dim_is_batch:
            raise NotImplementedError(
                "last_dim_is_batch=True not supported for GroupInvariantKernel."
            )

        x1_orbits = self.transformations(x1)  # Shape is ... x G x N x d
        G = x1_orbits.shape[-3]
        if self.is_isotropic:
            # Sum is over a single set of orbits
            # x2_orbits should be constructed by tiling x2 along the -3 axis
            dims = [-1] * (x2.dim() + 1)
            dims[-3] = G
            x2_orbits = x2.unsqueeze(-3).expand(dims)
            K_orbits = self.base_kernel.forward(x1_orbits, x2_orbits)
            K = torch.mean(K_orbits, dim=-3)
        else:
            # Sum is over all pairs of orbits
            # x2_orbits should be constructed by applying the transformations to x2
            x2_orbits = self.transformations(x2)  # Shape is ... x G x M x d

            if x2_orbits.shape[-3] != G or x1_orbits.shape[-3] != G:
                raise ValueError(
                    "Different numbers of orbits for x1 and x2. "
                    "Check that self.transformations returns a tensor of shape (..., G, N, d)."
                )

            # WARNING: This is quadratic in the number of orbits!
            # TODO: We should be able to parallelise this
            # TODO: Should be a way of doing it with less memory
            # Repeat each element of x1_orbits G times
            # New shape is G^2 x ... x N x d
            x1_orbits_expanded = x1_orbits.repeat_interleave(G, dim=-3)
            print(x1_orbits_expanded.shape)
            # Repeat the entire x2_orbits G times
            # New shape is ... x G^2 x M x d
            dims = [1] * x2_orbits.dim()
            dims[-3] = G
            x2_orbits_expanded = x2_orbits.repeat(dims)
            print(x2_orbits_expanded.shape)
            # Compute the kernel between each pair of expanded orbits = all combinations of orbits
            K_orbits = self.base_kernel.forward(x1_orbits_expanded, x2_orbits_expanded)
            print(K_orbits.shape)
            K = torch.mean(K_orbits, dim=-3)
            print(K.shape)

        if diag:
            return K.diag()
        else:
            return K



import torch
from gpytorch.kernels import MaternKernel, ScaleKernel

from invariantkernels import InvariantKernel, permutation_group


def compute_manual_kernel_matrix(k, x, y, transformations):
    N = x.shape[0]
    M = y.shape[0]
    G = transformations(x).shape[-3]
    manual_k_G_matrix = torch.zeros(N, M)
    for i in range(N):
        for j in range(M):
            # Get all permutations of x[i] and y[j]
            # These are tensors of shape (G, d)
            Gx_i = transformations(x[i].unsqueeze(0)).squeeze()
            Gy_j = transformations(y[j].unsqueeze(0)).squeeze()
            # Compute the kernel for all pairs of permutations
            for x_perm in Gx_i:
                for y_perm in Gy_j:
                    manual_k_G_matrix[i, j] += (
                        k(x_perm.unsqueeze(0), y_perm.unsqueeze(0)).to_dense().item()
                    )
    manual_k_G_matrix /= G**2
    return manual_k_G_matrix


def test_isotropic_group_invariant():
    n_datapoints = 10
    dimension = 3
    k = ScaleKernel(MaternKernel(nu=2.5))
    k_G = InvariantKernel(
        base_kernel=k,
        transformations=permutation_group,
        is_isotropic=True,
        is_group=True,
    )
    torch.manual_seed(0)

    x = torch.rand([n_datapoints, dimension])
    y = torch.rand([n_datapoints, dimension])

    k_G_matrix = k_G(x, y).to_dense()
    manual_k_G_matrix = compute_manual_kernel_matrix(k, x, y, permutation_group)
    assert torch.allclose(k_G_matrix, manual_k_G_matrix, atol=1e-5)


def test_anisotropic_group_invariant():
    n_datapoints = 10
    dimension = 3
    k = ScaleKernel(MaternKernel(nu=2.5, ard_num_dims=dimension))
    k_G = InvariantKernel(
        base_kernel=k,
        transformations=permutation_group,
        is_isotropic=False,
        is_group=True,
    )
    torch.manual_seed(0)

    x = torch.rand([n_datapoints, dimension])
    y = torch.rand([n_datapoints, dimension])

    k_G_matrix = k_G(x, y).to_dense()
    manual_k_G_matrix = compute_manual_kernel_matrix(k, x, y, permutation_group)
    assert torch.allclose(k_G_matrix, manual_k_G_matrix, atol=1e-5)


def test_batch_anisotropic_group_invariant():
    batch_size = 5
    n_datapoints = 10
    dimension = 3
    k = ScaleKernel(MaternKernel(nu=2.5, ard_num_dims=dimension))
    k_G = InvariantKernel(
        base_kernel=k,
        transformations=permutation_group,
        is_isotropic=False,
        is_group=True,
    )
    torch.manual_seed(0)

    x = torch.rand([batch_size, n_datapoints, dimension])
    y = torch.rand([batch_size, n_datapoints, dimension])

    k_G_matrix = k_G(x, y).to_dense()
    for i in range(batch_size):
        manual_k_G_matrix = compute_manual_kernel_matrix(
            k, x[i], y[i], permutation_group
        )
        assert torch.allclose(k_G_matrix[i], manual_k_G_matrix, atol=1e-5)
